package de.klausj.video.impl.gstreamer;

import java.awt.Component;
import java.nio.ByteBuffer;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;

import ips.media.event.MediaControlEventTransferAgent;

public class AudioSink {

	private static final boolean DEBUG=false;

	public native void configureVideoSink(ByteBuffer np,Component awtVideoComponent, boolean useJavaSound2);
	
	private MediaControlEventTransferAgent mceta;
	
	private Mixer.Info audioDeviceInfo;
//	private SourceDataLine sourceDataLine;
	private SourceDataLine sdl;
	
	private long audioFramePositionSync=0;
	private volatile long audioBytesWritten;
	private boolean videosinkSet=false; 
	
	private boolean useJavaSound=false;

	

	private ByteBuffer nativePointer;

	private native ByteBuffer init();
	private native void release(ByteBuffer nativePointer);
	    
	private volatile long audioFramePosition; 
	private volatile long audioTimeSync=-1;
	private Long audioStart=null;
	private int bufferSize;
	private int fakeBufFilled;
	private Long lastFakeBufferUpdate=null;
	private Long lastDelayCall=null;
	private float frameRate;
	private boolean enabled=true;
	private boolean active=false;

	public AudioSink() {
	    super();
	    nativePointer=init();
	}

	

	public boolean audioOpen(int sampleRate,int bytesPerSample,int channels,boolean bigEndian,int bufSize){
	    if(DEBUG){
	        System.out.println("Audio open");
	    }
		int sampleSizeInBits=bytesPerSample*8/channels;
		AudioFormat audioFormat=new AudioFormat(sampleRate, sampleSizeInBits, channels, true, bigEndian);
		if(DEBUG){
            System.out.println("Audio open "+audioFormat);
        }

		audioBytesWritten=0;
		audioFramePosition=0;
		audioFramePositionSync=0;
		lastDelayCall=null;
		if(!enabled){
		    bufferSize=bufSize;
		    fakeBufFilled=0;
		    lastFakeBufferUpdate=null;
		    frameRate=audioFormat.getFrameRate();
		    active=true;
		    return true;
		}
		if(audioDeviceInfo==null){
		    try {
                sdl=AudioSystem.getSourceDataLine(audioFormat);
            } catch (LineUnavailableException e) {
               return false;
            }
		}else{
		    try {
                AudioSystem.getSourceDataLine(audioFormat, audioDeviceInfo);
            } catch (LineUnavailableException e) {
               return false;
            }
		}
		if(sdl==null){
		    return false;
		}
		
			try {
			   
				if(DEBUG)System.out.println("Audio "+audioFormat+" Pref Buffer size "+bufSize);
				sdl.open(audioFormat,bufSize);
				bufferSize=sdl.getBufferSize();
				if(DEBUG)System.out.println("Buffer size "+bufferSize+" "+bufferSize/(audioFormat.getFrameSize()*audioFormat.getSampleRate()));
				 if(!sdl.isActive()){
			            sdl.start();
			            if(DEBUG)System.out.println("SDL started");
			      }
				 return true;
			} catch (LineUnavailableException e) {
				
				return false;
			}
		
	}
	
	private void updateFakeBuffer(){
	    if(lastFakeBufferUpdate==null){
	        lastFakeBufferUpdate=System.currentTimeMillis();
	    }
	    long bufPlayedTime=System.currentTimeMillis()-lastFakeBufferUpdate;
	    int bytesPlayed=(int)(frameRate*bufPlayedTime/1000);
	    int newFbf=fakeBufFilled-bytesPlayed;
	    if(newFbf<0){
	        newFbf=0;
	    }
	    fakeBufFilled=newFbf;
	    lastFakeBufferUpdate=System.currentTimeMillis();
	}
	
	
	public int audioWrite(byte[] buf){
	    if (DEBUG){
	        int filled=sdl.getBufferSize()-sdl.available();
	        // check for null buffer
	        boolean isNull=true;
	        for(int i=0;i<buf.length;i++){
	            if(buf[i]!=0){
	                isNull=false;
	                break;
	            }
	        }
	        if(isNull){
	            System.out.println("Silence buffer!");
	        }
	        System.out.println("Audio filled/write "+filled+"/"+buf.length+" bytes.");
	     
	        
	    }
		
		
//		while(audioBytesDelay()>8000){
//			try {
//				Thread.sleep(50);
//			} catch (InterruptedException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}
//		}
//	    int written=0;
//		int avail=sdl.available();
//		int c=0;
//		while(avail<buf.length && c<40){
//		    try {
//                Thread.sleep(50);
//                c++;
//            } catch (InterruptedException e) {
//                // TODO Auto-generated catch block
//                e.printStackTrace();
//            }
//		    avail=sdl.available();
//		}
	    int written=0;
	    if(enabled){
	        int avail=sdl.available();
//	        if(avail==0)return 0;
	        int toWrite=avail;
	        if(buf.length<toWrite){
	            toWrite=buf.length;
	        }
	        if(toWrite==0){
	            try {
                    Thread.sleep(20);
                } catch (InterruptedException e) {
                   // OK
                }
	        }else{
	            // write as much as possible
	            // blocking until the whole buffer can be written causes choppy video!
	            written=sdl.write(buf, 0, toWrite);
	        }
	    }else{
	        if(audioStart==null){
                audioStart=System.currentTimeMillis();
            }
	        do{
	        updateFakeBuffer();
	        int avail=bufferSize-fakeBufFilled;
	        if(avail>=buf.length){
	            written=buf.length;
	            fakeBufFilled+=written;
	            
	        }else{
	            try {
                    Thread.sleep(50);
                } catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
	        }
	        }while(written!=buf.length && active);
	      
	        
	    }
//		written=buf.length;
		audioBytesWritten+=written;
		
//		if(!sdl.isActive() && sdl.getBufferSize()-sdl.available()>=512){
//		int filled=sdl.getBufferSize()-sdl.available();
		 if(DEBUG)System.out.println("Audio written "+written+" bytes");
		return written;
	}
	
	public int audioDelay(){
	    AudioFormat af=sdl.getFormat();
	    
        int frameSize=af.getFrameSize();
//        
//        Integer framesBySysClock=null;
//	    if(lastDelayCall==null){
//	        lastDelayCall=System.currentTimeMillis();
//	                
//	    }
//		if(audioStart==null)return 0;
//		long playTime=System.currentTimeMillis()-audioStart;
//	
//		long framesPlayed=playTime*48;
//		long bytesPlayed=framesPlayed*4;
////		long bytesWritten =audioFramesWritten*4;
//		long delay= audioFramesWritten-bytesPlayed;
//		if(delay<0){
//			delay=0;
//		}
////		System.out.println("Delay: "+delay+" "+playTime/1000);
//		return delay;
		
//		float frameRate=af.getFrameRate();
//		int sSize=af.getFrameSize()/af.getChannels();
		int bufSize=sdl.getBufferSize();
		int avail=sdl.available();
//		int availBlocks=avail/1920;
		int filled=bufSize-(avail);
//		int filledBlocks=filled/1920;
//		int filledBlockwise=filledBlocks*1920;
		// samples (frames?) queued
		int delay;
//		delay=filled/frameSize;
		long retFramePos;
		if(enabled){
//		    long framePos=sdl.getLongFramePosition();
//		    retFramePos=framePos-audioFramePositionSync;
		}else{
		    if(audioStart==null){
		        retFramePos=0;
		    }else{
		        float frameRate=af.getFrameRate();
		        long playTime=System.currentTimeMillis()-audioStart;
		        retFramePos=(long)(playTime*frameRate/1000);
		    }
		}
		
//		if(framePos!=audioFramePosition){
////		    System.out.println("delay: framepos changed: "+framePos);
//		    audioTimeSync=System.currentTimeMillis();
//		}else{
//		   
//		    if(audioTimeSync!=-1){
//		        long now=System.currentTimeMillis();
//		        long timeDiff=now-audioTimeSync;
//		        long addFrames=(long)((timeDiff*frameRate)/1000.0);
//		        retFramePos=framePos+addFrames;
////		        System.out.println("delay: framepos add frames: "+addFrames);
//		    }else{
////		        System.out.println("delay: framepos no change");
//		    }
//		}
//		audioFramePosition=framePos;
//		long delay=(bufSize-avail)/frameSize;
//		delay=(int) (audioBytesWritten/frameSize -retFramePos);
		
//		long playedFrames=(audioBytesWritten-filled)/frameSize;
//		long playedMilliSeconds=(long)(((float)playedFrames*frameRate)*1000);
		delay=(filled/frameSize);
		
		if(DEBUG) System.out.println("Delay: "+delay);
		return delay;
	}
	
	public void audioReset(){
	    if(DEBUG)System.out.println("Audio reset...");
	    
	    if(!enabled){
	        audioStart=null;
	        active=false;
	    }else{
		if(sdl!=null){
		    sdl.stop();
			sdl.flush();
			 if(DEBUG)System.out.println("Audio flushed.");
			 // sync
			 audioFramePositionSync=sdl.getLongFramePosition();
			 audioBytesWritten=0;
			sdl.start();
			if(DEBUG)System.out.println("Audio restarted.");
		}
	    }
	}
	
	public boolean audioUnprepare(){
	    
	    return  true;
	}
	
	public boolean audioClose(){
		if(DEBUG)System.out.println("Audio close...");
		if(enabled && sdl!=null){
			sdl.stop();
			sdl.close();
		}
		audioStart=null;
		active=false;
		audioBytesWritten=0;
		audioFramePositionSync=0;
		if(DEBUG)System.out.println("Audio closed.");
		return true;

	}

	protected void finalize() {
		release(nativePointer);
		System.out.println("audio sink finalized");
	}



  



    public Mixer.Info getAudioDeviceInfo() {
        return audioDeviceInfo;
    }

    public void setAudioDeviceInfo(Mixer.Info audioDeviceInfo) {
        this.audioDeviceInfo = audioDeviceInfo;
    }
    public ByteBuffer nativePointer() {
      
        return nativePointer;
    }


}
